# Copyright (C) 2016 Sebastian Pipping <sebastian@pipping.org>
# Licensed under GNU Affero GPL v3 or later

import cPickle as pickle
import datetime
from StringIO import StringIO

from unittest import TestCase

from jawanndenn.poll import _Poll, _PICKLE_POLL_VERSION


def _create_v1_poll():
    p = _Poll()
    p._version = 1
    del p.__dict__['_created_at']
    del p.__dict__['_equal_width']
    return p


def _save_and_load(poll):
    f = StringIO()
    pickle.dump(poll, f)
    f.seek(0)
    return pickle.load(f)


class UpgradeTest(TestCase):
    def test_upgrade_1_to_latest(self):
        v1 = _create_v1_poll()

        self.assertEquals(v1._version, 1)
        self.assertFalse(hasattr(v1, '_created_at'))
        self.assertFalse(hasattr(v1, '_equal_width'))

        latest = _save_and_load(v1)

        self.assertEquals(latest._version, _PICKLE_POLL_VERSION)
        self.assertTrue(isinstance(latest._created_at, datetime.datetime))
        self.assertTrue(isinstance(latest._equal_width, bool))



# Copyright (C) 2016 Sebastian Pipping <sebastian@pipping.org>
# Licensed under GNU Affero GPL v3 or later

import cPickle as pickle
import datetime
import hashlib
import logging
import os
import shutil
import tempfile

from threading import Lock

from jawanndenn.markup import safe_html


DEFAULT_MAX_POLLS = 1000
DEFAULT_MAX_VOTER_PER_POLL = 40

_MAX_POLLS = DEFAULT_MAX_POLLS
_MAX_VOTERS_PER_POLL = DEFAULT_MAX_VOTER_PER_POLL

_KEY_OPTIONS = 'options'
_KEY_TITLE = 'title'
_KEY_EQUAL_WIDTH = 'equal_width'

_PICKLE_PROTOCOL_VERSION = 2

_PICKLE_CONTENT_VERSION = 1
_PICKLE_POLL_VERSION = 3
_PICKLE_POLL_DATABASE_VERSION = 1


_log = logging.getLogger(__name__)


def _get_random_sha256():
    return hashlib.sha256(os.urandom(256 / 8)).hexdigest()


def apply_limits(polls, votes_per_poll):
    global _MAX_POLLS
    global _MAX_VOTERS_PER_POLL
    _MAX_POLLS = polls
    _MAX_VOTERS_PER_POLL = votes_per_poll


class _Poll(object):
    def __init__(self):
        # Version 1
        self.config = {}
        self.votes = []
        self._lock = Lock()
        self._version = _PICKLE_POLL_VERSION

        # Version 2 and later
        self._created_at = datetime.datetime.now()

        # Version 3 and later
        self._equal_width = False

    def __getstate__(self):
        d = self.__dict__.copy()
        del d['_lock']
        return d

    def __setstate__(self, d):
        self.__dict__.update(d)
        self._lock = Lock()

        initial_version = self._version

        if self._version == 1:
            self._version = 2
            self._created_at = datetime.datetime.now()

        if self._version == 2:
            self._version = 3
            self._equal_width = False

        assert self._version == _PICKLE_POLL_VERSION

        if self._version != initial_version:
            _log.debug('Upgraded poll from version %d to version %d'
                    % (initial_version, self._version))

    @staticmethod
    def from_config(config):
        poll = _Poll()

        if _KEY_OPTIONS not in config \
                or _KEY_TITLE not in config:
            raise ValueError('Malformed configuration: %s' % config)

        poll.config = {
            _KEY_EQUAL_WIDTH: bool(config.get(_KEY_EQUAL_WIDTH, False)),
            _KEY_TITLE: safe_html(config[_KEY_TITLE]),
            _KEY_OPTIONS: map(safe_html, config[_KEY_OPTIONS]),
        }
        return poll

    @property
    def options(self):
        return self.config[_KEY_OPTIONS]

    def vote(self, person, votes):
        with self._lock:
            if len(self.votes) >= _MAX_VOTERS_PER_POLL:
                raise ValueError('Too many votes per poll')
            if len(votes) != len(self.options):
                raise ValueError('Malformed vote')
            self.votes.append((safe_html(person), votes))


class PollDatabase(object):
    def __init__(self):
        self._db = {}
        self._db_lock = Lock()
        self._version = _PICKLE_POLL_DATABASE_VERSION

    def __getstate__(self):
        d = self.__dict__.copy()
        del d['_db_lock']
        return d

    def __setstate__(self, d):
        self.__dict__.update(d)
        self._db_lock = Lock()

    def add(self, config):
        poll = _Poll.from_config(config)
        poll_id = _get_random_sha256()

        with self._db_lock:
            if len(self._db) >= _MAX_POLLS:
                raise ValueError('Too many polls')
            if poll_id in self._db:
                raise ValueError('ID collision: %s' % poll_id)
            self._db[poll_id] = poll

        return poll_id

    def get(self, poll_id):
        with self._db_lock:
            return self._db[poll_id]

    def load(self, filename):
        with open(filename, 'rb') as f:
            d = pickle.load(f)

        if d['version'] != _PICKLE_CONTENT_VERSION:
            raise ValueError('Content version mismatch')

        self.__dict__.update(d['data'].__dict__)
        _log.info('%d polls loaded.' % len(self._db))

    def save(self, filename):
        with self._db_lock:
            d = {
                'version': _PICKLE_CONTENT_VERSION,
                'data': self,
            }

            fd, tempfilename = tempfile.mkstemp(
                    dir=os.path.dirname(filename),
                    prefix='%s-tmp' % os.path.basename(
                            filename.replace('.pickle', '')),
                    suffix='.pickle',
                    )

            with os.fdopen(fd, 'w') as f:
                pickle.dump(d, f, _PICKLE_PROTOCOL_VERSION)

            shutil.move(tempfilename, filename)

            _log.info('%d polls saved.' % len(self._db))



# Copyright (C) 2016 Sebastian Pipping <sebastian@pipping.org>
# Licensed under GNU Affero GPL v3 or later

import argparse
import errno
import logging
import os
import sys

from jawanndenn.poll import (apply_limits,
        DEFAULT_MAX_POLLS, DEFAULT_MAX_VOTER_PER_POLL)


_BOTTLE_BACKENDS = (
#   'cgi',  # see https://github.com/bottlepy/bottle/issues/836
#   'flup',
    'gae',
    'wsgiref',
#   'cherrypy',
    'paste',
#   'rocket',
    'waitress',
    'gunicorn',
    'eventlet',
    'gevent',
    'diesel',
#   'fapws3',  # symptom "XML Parsing Error: no root element found"
    'tornado',
    'twisted',
    'meinheld',
    'bjoern',
    'auto',
)

_DEFAULT_BACKEND = 'paste'

_log = logging.getLogger(__name__)


def _require_hash_randomization():
    if not sys.flags.hash_randomization:
        _log.info('Hash randomization found to be disabled.')
        if os.environ.get('PYTHONHASHSEED') == 'random':
            _log.error('Unexpected re-execution loop detected, shutting down.')
            sys.exit(1)

        _log.info('Re-executing with hash randomization enabled...')
        env = os.environ.copy()
        env['PYTHONHASHSEED'] = 'random'
        argv = [sys.executable] + sys.argv
        os.execve(argv[0], argv, env)


def main():
    parser = argparse.ArgumentParser(prog='jawanndenn')
    parser.add_argument('--debug', action='store_true',
            help='Enable debug mode (default: disabled)')
    parser.add_argument('--host', default='127.0.0.1', metavar='HOST',
            help='Hostname or IP address to listen at (default: %(default)s)')
    parser.add_argument('--port', default=8080, type=int, metavar='PORT',
            help='Port to listen at (default: %(default)s)')
    parser.add_argument('--database-pickle', default='~/jawanndenn.pickle', metavar='FILE',
            help='File to write the database to (default: %(default)s)')
    parser.add_argument('--server', default='paste', metavar='BACKEND',
            help='bottle backend to use (default: %%(default)s)'
                '; as of this writing bottle supports: %s. '
                'For the most current list, please check the documentation '
                'of bottle.'
                % ', '.join(sorted(b for b in _BOTTLE_BACKENDS
                                   if b != _DEFAULT_BACKEND))
                )

    limits = parser.add_argument_group('limit configuration')
    limits.add_argument('--max-polls', type=int, metavar='COUNT',
            default=DEFAULT_MAX_POLLS,
            help='Maximum number of polls total (default: %(default)s)')
    limits.add_argument('--max-votes-per-poll', type=int, metavar='COUNT',
            default=DEFAULT_MAX_VOTER_PER_POLL,
            help='Maximum number of votes per poll (default: %(default)s)')

    options = parser.parse_args()

    logging.basicConfig(level=logging.DEBUG if options.debug else logging.INFO)

    _require_hash_randomization()

    # NOTE: gevent patching needs to happen before importing bottle
    if options.server == 'gevent':
        from gevent.monkey import patch_all
        patch_all()

    # Heavy imports are down here to keep --help fast
    from jawanndenn.app import db, run_server, STATIC_HOME_LOCAL

    apply_limits(
        polls=options.max_polls,
        votes_per_poll=options.max_votes_per_poll,
    )

    _log.debug('Serving static files from "%s"' % STATIC_HOME_LOCAL)

    filename = os.path.expanduser(options.database_pickle)

    try:
        db.load(filename)
    except IOError as e:
        if e.errno != errno.ENOENT:
            raise
        db.save(filename)  # catch saving trouble early

    try:
        run_server(options)
    finally:
        db.save(filename)


if __name__ == '__main__':
    main()



# Copyright (C) 2016 Sebastian Pipping <sebastian@pipping.org>
# Licensed under GNU Affero GPL v3 or later

_REPLACEMENTS_IN_ORDER = (
    ('**', '<strong>', '</strong>'),
    ('*', '<em>', '</em>'),
    ('__', '<strong>', '</strong>'),
    ('_', '<em>', '</em>'),
    ('`', '<tt>', '</tt>'),
)

_CLOSING_OF = {
        prefix: closing
        for prefix, _, closing
        in _REPLACEMENTS_IN_ORDER
}


def safe_html(text):
    if not isinstance(text, basestring):
        raise ValueError('Not a string: %s' % text)

    # KEEP IN SYNC with javascript client side!

    text = text.replace('&', '&amp;') \
            .replace('<', '&lt;') \
            .replace('>', '&gt;')

    chunks = []

    opened = []
    while text:
        for prefix, opening, closing in _REPLACEMENTS_IN_ORDER:
            if text.startswith(prefix):
                if opened and opened[-1] == prefix:
                    # Close tag
                    chunks.append(closing)
                    opened.pop()
                else:
                    # Open tag
                    chunks.append(opening)
                    opened.append(prefix)

                text = text[len(prefix):]
                break
        else:
            chunks.append(text[0])
            text = text[1:]

    # Close all unclosed tags
    for prefix in reversed(opened):
        chunks.append(_CLOSING_OF[prefix])

    return ''.join(chunks)



# Copyright (C) 2016 Sebastian Pipping <sebastian@pipping.org>
# Licensed under GNU Affero GPL v3 or later

import json
import logging
import os
import sys

import bottle
import pkg_resources

from jawanndenn.metadata import APP_NAME
from jawanndenn.poll import PollDatabase


_log = logging.getLogger(__name__)


STATIC_HOME_LOCAL = os.path.abspath(os.path.normpath(
        os.path.join(os.path.dirname(__file__), 'static')
            if os.path.exists(os.path.join(os.path.dirname(__file__),
                    '..', 'setup.py')) else
            pkg_resources.resource_filename(APP_NAME, 'static')
        ))
_STATIC_HOME_REMOTE = '/static'


db = PollDatabase()


def _to_json(e):
    return json.dumps(e)


@bottle.get('/static/<path:path>')
def _static(path):
    content_type = {
                'css': 'text/css',
                'js': 'application/javascript',
                'xhtml': 'application/xhtml+xml',
            }[path.split('.')[-1]]
    bottle.response.content_type = content_type
    return bottle.static_file(path, root=STATIC_HOME_LOCAL)


@bottle.get('/')
def _index():
    bottle.response.content_type = 'application/xhtml+xml'
    return bottle.static_file('html/setup.xhtml', root=STATIC_HOME_LOCAL)


@bottle.post('/create')
def _create():
    config = json.loads(bottle.request.forms['config'])
    poll_id = db.add(config)
    bottle.redirect('/poll/%s' % poll_id)


@bottle.get('/poll/<poll_id>')
def _poll(poll_id):
    db.get(poll_id)
    bottle.response.content_type = 'application/xhtml+xml'
    return bottle.static_file('html/poll.xhtml', root=STATIC_HOME_LOCAL)


@bottle.get('/data/<poll_id>')
def _data(poll_id):
    poll = db.get(poll_id)
    return _to_json({
        'config': poll.config,
        'votes': poll.votes,
    })


@bottle.post('/vote/<poll_id>')
def _vote(poll_id):
    voterName = bottle.request.forms['voterName']
    poll = db.get(poll_id)
    votes = [bottle.request.forms.get('option%d' % i) == 'on'
            for i in xrange(len(poll.options))]
    poll.vote(voterName, votes)

    bottle.redirect('/poll/%s' % poll_id)


def run_server(options):
    if options.debug:
        bottle.debug(True)

    try:
        bottle.run(
                host=options.host,
                port=options.port,
                server=options.server,
                )
    except ImportError as e:
        if (options.server == 'cherrypy' and
                e.args == ('cannot import name wsgiserver',)):
            message = 'Bottle does not support CherryPy >=9.0.0, properly.'
        elif (options.server == 'flup' and
                e.args == ('No module named _dummy_thread',)):
            message = 'flup is broken and not supported.'
        elif (options.server == 'rocket' and
                e.args == ('No module named methods.wsgi',)):
            message = 'Rocket is broken and not supported.'
        else:
            message = ('WSGI server "%s" does not seem to be available.'
                       % options.server)

        _log.error(message)
        sys.exit(2)
